import numpy as np
import scipy.stats
from typing import Tuple
from ..base import MixtureModel
from ...utils.exceptions import NotEnoughVariation
from ...utils.static import Params, eps


class CauchyMixtureModel(MixtureModel):
    _param_signature_ = {'loc': {'type': float, 'max': np.inf, 'min': -np.inf},
                         'scale':{'type': float, 'max': np.inf, 'min': 0},}
    _support = {'min': -np.inf, 'max': np.inf, 'dtype': float}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = 'CauchyMixtureModel'
        self.zero_truncated = True
        self.discrete = False
        self.n_variables = 1

    def likelihood(self, weights, data, log=False, **params):
        likelihoods = scipy.stats.cauchy.pdf(data, **params)
        likelihood_weighted = likelihoods * weights
        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def _fit_deterministic_(self, X):
        loc, scale = scipy.stats.cauchy.fit(X)
        return {'loc': loc[..., np.newaxis], 'scale': scale[..., np.newaxis]}, np.array([1])

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        # https://www.mdpi.com/2073-8994/11/9/1186/pdf
        X = X.copy()
        X.sort()

        weights = np.ones(n_components) / n_components
        locs = np.quantile(X, (np.arange(1, n_components + 1) / (n_components + 1)))
        locs *= ((np.random.random(n_components) + 99.5) / 100) # force some variation in case quantiles are identical
        scales = np.repeat(.5 * scipy.stats.iqr(X), n_components)
        if scales[0] == 0:
            raise NotEnoughVariation(model_type=self.name)

        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=-1)
        log_likelihoods = [-np.inf]
        for iteration in range(self.max_iterations):

            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False, loc=locs, scale=scales)
            r = (eps + likelihood_weighted) / (likelihood_weighted + eps).sum(axis=1)[..., np.newaxis]

            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)

            if self.check_convergence(log_likelihoods, ignore_assertion=True):
                break

            # Maximization
            component_n = r.sum(axis=0).round().astype(int)
            if component_n.min() <= 5:
                break
            loc_list, scale_list = [],[]
            previous_component_end = 0
            for component_size in component_n:
                X_component = X[previous_component_end:component_size+previous_component_end]
                ixs = np.array((component_size/4, component_size/2, 3*component_size/4)).round().astype(int)
                q1, loc, q3 = X_component[ixs]
                scale_list.append((q3 - q1) / 2)
                loc_list.append(loc)
                previous_component_end += component_size
            locs = np.array(loc_list)
            scales = np.array(scale_list)
            weights = component_n / component_n.sum()
        return {'loc': locs, 'scale': scales}, weights, log_likelihoods[-1]

    def _generate_component(self, component, n):
        return scipy.stats.cauchy.rvs(loc=self.params['loc'][component], scale=self.params['scale'][component], size=n)

    def _cdf_separate(self, X):
        return scipy.stats.cauchy.cdf(X, **self.params)

    def _sf_separate(self, X):
        return scipy.stats.cauchy.sf(X, **self.params)


import numpy as np
import scipy.stats
from typing import Tuple, Union
from ..base import MixtureModel
from ...utils.static import Params, eps


class ExponentialMixtureModel(MixtureModel):
    _param_signature_ = {'scale': {'type': float, 'max': np.inf, 'min': 0}}
    _support = {'min': 0-eps, 'max': np.inf, 'dtype': float}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "ExponentialMixtureModel"
        self.n_variables = 1
        self.zero_truncated = True

    def likelihood(self, weights: np.ndarray, data: np.ndarray, log=False, **params) -> Union[
        np.ndarray, Union[float, np.float64]]:
        # **params are scale= 1 / lambda
        likelihood = scipy.stats.expon.pdf(data, **params)
        likelihood_weighted = likelihood * weights
        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def _fit_deterministic_(self, X: np.ndarray) -> Tuple[Params, np.ndarray]:
        # https://en.wikipedia.org/wiki/Exponential_distribution#Parameter_estimation
        lambd = 1 / X.mean()  # MLE estimate
        lambd = lambd - lambd / X.shape[0]  # bias corrected MLE estimate
        scale = 1 / lambd  # scipy parameterization uses loc and scale
        weights = np.array([1])
        return {'scale': scale.reshape(-1)}, weights

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:

        weights = np.ones(n_components) / n_components
        lambdas = 1 / (X.mean() * (np.random.random(n_components) + .5))
        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=1)
        log_likelihoods = [-np.inf]

        for iteration in range(self.max_iterations):
            # calculate the maximum likelihood of each observation xi

            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False, scale=1 / lambdas)
            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)
            if self.check_convergence(log_likelihoods):
                break
            r = (likelihood_weighted + eps) / (likelihood_weighted + self.eps).sum(axis=1)[..., np.newaxis]

            # Maximization
            lambdas = 1 / np.average(X_rep, weights=r, axis=0)
            weights = r.sum(axis=0) / r.sum()

        return {'scale': 1 / lambdas}, weights, log_likelihoods[-1]

    def _generate_component(self, component, n):
        return scipy.stats.expon.rvs(scale=self.params['scale'][component], size=n)

    def _cdf_separate(self, X):
        return scipy.stats.expon.cdf(X, **self.params)

    def _sf_separate(self, X):
        return scipy.stats.expon.sf(X, **self.params)


import numpy as np
import scipy.stats
from typing import Tuple
from ..base import MixtureModel
from ...utils.static import Params
import warnings
from smm import SMM


class StudentsTMixtureModel(MixtureModel):
    _param_signature_ = {'loc': {'type': float, 'max': np.inf, 'min': -np.inf},
                         'scale': {'type': float, 'max': np.inf, 'min': 0},
                         'df': {'type': float, 'max': np.inf, 'min': 0},}
    _support = {'min': -np.inf, 'max': np.inf, 'dtype': float}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = 'StudentsTMixtureModel'
        self.n_iter = 1
        self.zero_truncated = True
        self.discrete = False
        self.n_variables = 1

    def likelihood(self, weights, data, log=False, **params):
        likelihoods = scipy.stats.t.pdf(data, **params)
        likelihood_weighted = likelihoods * weights
        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def _fit_deterministic_(self, X):
        df, loc, scale = scipy.stats.t.fit(X)
        return {'df': df[...,np.newaxis], 'loc': loc[...,np.newaxis], 'scale': scale[...,np.newaxis]}, np.array([1])

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        if len(X.shape) == 1:
            X = X[..., np.newaxis]
        model = SMM(n_components)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            model.fit(X)
        loc = model.means_.reshape(-1)
        scale = np.sqrt(model.covars_.reshape(-1))
        df = model.degrees_
        weights = model.weights_
        X_rep = np.repeat(X, n_components, axis=1)
        log_likelihood = self.likelihood(weights=weights, data=X_rep, log=True, loc=loc, scale=scale, df=df)
        return {'loc': loc, 'scale': scale, 'df': df}, weights, log_likelihood

    def _generate_component(self, component, n):
        return scipy.stats.t.rvs(df=self.params['df'][component], scale=self.params['scale'][component],
                                 loc=self.params['loc'][component], size=n)

    def _cdf_separate(self, X):
        return scipy.stats.t.cdf(X, **self.params)

    def _sf_separate(self, X):
        return scipy.stats.t.sf(X, **self.params)


import numpy as np
import scipy.stats
from typing import Tuple, Union
from ..base import MixtureModel
from ...utils.static import Params, eps
from scipy.special import zeta


def zeta_prime(x: np.ndarray) -> np.ndarray:
    h = 1e-6
    return (zeta(x + h, 1) - zeta(x - h, 1)) / (2 * h)


def fit_zeta_mle(d: np.ndarray) -> np.ndarray:
    x = np.linspace(1.01, 100, num=10000)
    critical = - zeta_prime(x) / zeta(x)
    return x[np.argmin(np.abs((critical - (1 / d.shape[0]) * np.log(d).sum())))]


class ZetaMixtureModel(MixtureModel):
    _param_signature_ = {'a': {'type': float, 'max': np.inf, 'min': 1},}
    _support = {'min': 1-eps, 'max': np.inf, 'dtype': int}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = 'ZetaMixtureModel'
        self.min_components = 1
        self.discrete = True
        self.n_variables = 1

    def _fit_deterministic_(self, X: np.ndarray):
        a = fit_zeta_mle(X)
        return {'a': a[..., np.newaxis]}, np.array([1])

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        weights = np.ones(n_components) / n_components
        a = np.array((1.01 * np.ones(n_components),
                      np.random.uniform(.5, 1.5, size=n_components) * fit_zeta_mle(X))).max(axis=0)
        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=-1)
        log_likelihoods = [-np.inf]
        for iteration in range(self.max_iterations):
            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False, a=a)
            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)

            if self.check_convergence(log_likelihoods, ignore_assertion=True):
                break

            r = (eps + likelihood_weighted) / (likelihood_weighted + eps).sum(axis=1)[..., np.newaxis]

            a = list()
            chosen = list()
            data_available = np.ones(X.shape[0])
            for i in range(n_components - 1):
                p = r[:, i]
                mask = np.random.binomial(n=1, p=p) # randomly choose pieces according to weight
                actual_mask = (data_available * mask).astype(bool)
                data_chosen = X[actual_mask]
                chosen.append(data_chosen)
                data_available *= (1 - mask)
                a.append(fit_zeta_mle(data_chosen))
            chosen.append(X[data_available.astype(bool)])
            a.append(fit_zeta_mle(X[data_available.astype(bool)]))
            weights = r.mean(axis=0)

        return {'a': a}, weights, log_likelihoods[-1]

    def likelihood(self, weights: np.ndarray, data: np.ndarray, log=False, **params) -> Union[np.ndarray, Union[float, np.float64]]:
        likelihood = scipy.stats.zipf.pmf(data, **params)
        likelihood_weighted = likelihood * weights
        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def _generate_component(self, component, n):
        return scipy.stats.zipf.rvs(a=self.params['a'][component], size=n)

    def _cdf_separate(self, X):
        return scipy.stats.zipf.cdf(X + self.eps, **self.params)

    def _sf_separate(self, X):
        return scipy.stats.zipf.sf(X - self.eps, **self.params)




import numpy as np
import scipy.stats
from typing import Tuple, Union
from ..base import MixtureModel
from ...utils.static import Params, eps

class GeometricMixtureModel(MixtureModel):
    _param_signature_ = {'p': {'type': float, 'max': 1+eps, 'min': 0}, }
    _support = {'min': 0, 'max': np.inf, 'dtype': int}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = 'GeometricMixtureModel'
        self.discrete = True
        self.n_variables = 1

    def _fit_deterministic_(self, X: np.ndarray) -> Tuple[Params, np.ndarray]:
        p = 1 / X.mean(axis=0)
        return {'p': p[..., np.newaxis]}, np.array([1])

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        weights = np.ones(n_components) / n_components
        p = (1 / X.mean(axis=0) * np.random.uniform(.25, 2, n_components))
        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=-1)
        log_likelihoods = [-np.inf]

        for iteration in range(self.max_iterations):
            # calculate the maximum likelihood of each observation xi

            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False, p=p)
            r = (likelihood_weighted + eps) / (likelihood_weighted + eps).sum(axis=1)[..., np.newaxis]

            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)

            if self.check_convergence(log_likelihoods):
                break
            # Maximization
            p = 1 / np.average(X_rep, weights=r, axis=0)
            weights = r.sum(axis=0) / r.sum()

        return {'p': p}, weights, log_likelihoods[-1]

    def likelihood(self, weights: np.ndarray, data: np.ndarray, log=False, **params) -> Union[
        np.ndarray, Union[float, np.float64]]:
        likelihoods = scipy.stats.geom.pmf(data, **params)
        likelihood_weighted = likelihoods * weights
        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def _generate_component(self, component, n):
        return scipy.stats.geom.rvs(p=self.params['p'][component], size=n)

    def _cdf_separate(self, X):
        return scipy.stats.geom.cdf(X+self.eps, **self.params)

    def _sf_separate(self, X):
        return scipy.stats.geom.sf(X-self.eps, **self.params)


from typing import Dict
import numpy as np
from scipy.sparse import random as sparse_random
from .exceptions import ModelNotFit


def jiggle_likelihoods(likelihoods):
    m, n = likelihoods.shape
    return likelihoods + 1e-8 * np.random.random(size=likelihoods.shape) * sparse_random(m=m, n=n, density=.01).A


def transpose_params(params):
    params = params.copy()
    for p in params:
        params[p] = params[p].T
    return params


def sort_params(weights: np.ndarray, params: Dict, zeros=False):
    if params is None:
        return weights, params
    if zeros:
        zeroweight = weights[0]
        weights = weights[1:]

    order = weights.argsort()[::-1]
    params = params.copy()
    for name in params:
        if isinstance(params[name], dict):
            for subname in params[name]:
                params[name][subname] = params[name][subname][order]
        else:
            params[name] = params[name][order]

    if zeros:
        weights = np.hstack((zeroweight, weights[order]))
    else:
        weights = weights[order]
    return weights, params


def simplify_params(params):
    """
        replace numpy arrays with lists, for jsonification handling
    :param params:
    :return:
    """
    params = params.copy()
    for key in params:
        if isinstance(params[key], np.ndarray):
            params[key] = params[key].tolist()
        elif isinstance(params[key], dict):
            for subkey in params[key]:
                if isinstance(params[key][subkey], np.ndarray):
                    params[key][subkey] = params[key][subkey].tolist()
    return params


def checkfit(func):
    def checker(self, *args, **kwargs):
        if self.isfit:
            return func(self, *args, **kwargs)
        else:
            raise ModelNotFit(f'{self.name} Not yet fit. Attempted function: {func.__name__}', self.name)
    return checker




import numpy as np
import scipy.stats
from typing import Tuple, Union
from ..base import MixtureModel
from ...utils.static import Params, eps

BASE_N = 100

class ZeroInflatedBinomialMixtureModel(MixtureModel):
    _param_signature_ = {'p': {'type': float, 'max': 1 + eps, 'min': 0},}
    _support = {'min': 0 - eps, 'max': np.inf, 'dtype': int}
    def __init__(self, med_abs_dev=False, standardize=False, **kwargs):
        super().__init__(**kwargs)
        self.name = 'ZeroInflatedBinomialMixtureModel'
        self.skip_deterministic = True
        self.min_components = 2
        self.standardize = standardize
        self.zero_inflated = True
        self.discrete = True
        self.n_variables = 2

    def _fit_deterministic(self, X: np.ndarray):
        return self._fit_(X, self.min_components)

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        assert n_components > 1
        weights = np.ones(n_components) / n_components
        ps = (X[:, 0].sum() / X[:, 1].sum()) * np.random.uniform(.5, 1.5, size=n_components - 1)
        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=-1)
        log_likelihoods = [-np.inf]
        for iteration in range(self.max_iterations):
            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False, p=ps)
            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)

            if self.check_convergence(log_likelihoods):
                break

            r = (eps + likelihood_weighted) / (likelihood_weighted + eps).sum(axis=1)[..., np.newaxis]

            # maximization
            ps = np.average(X_rep[:, 0, 1:] / X_rep[:, 1, 1:], axis=0, weights=r[:, 1:])
            weights = r.mean(axis=0)

        return {'p': ps}, weights, log_likelihoods[-1]

    def likelihood(self, weights: np.ndarray, data: np.ndarray, log=False, **params) -> Union[
        np.ndarray, Union[float, np.float64]]:
        assert weights.shape[0] > 1
        likelihood_zero = (data[:, 0, 0] == 0).astype(np.int64)[:, np.newaxis]
        likelihood_binom = scipy.stats.binom.pmf(n=data[:, 1, 1:], k=data[:, 0, 1:], **params)
        likelihood_weighted = np.concatenate((likelihood_zero, likelihood_binom), axis=1) * weights
        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def generate_zero_inflated_component(self, n):
        zero_data, zero_labels = super().generate_zero_inflated_component(n)
        zero_data = self.add_n_to_k(zero_data)
        return zero_data, zero_labels

    def _generate_component(self, component, n, k):
        component -=1

        if isinstance(k, int):
            k = np.array(k)
        elif callable(k):
            k = k(n)

        if len(k.shape) > 1:
            if k.shape[1] == 1:
                k = k.flatten()
            else:
                raise Exception(f'Dont know how to handle k of shape {k.shape}')
        elif len(k.shape) < 1:
            k = k[np.newaxis]

        if k.shape[0] == self.n_components - 1:
            k = k[component]

        if k.shape == (n,):
            pass
        elif k.shape in [(1,), ()]:
            k = np.repeat(k, repeats=n)
        else:
            raise Exception(f'Dont know how to handle k of shape: {k.shape}')

        return np.stack((scipy.stats.binom.rvs(n=k, p=self.params['p'][component], size=n), k)).T

    def add_n_to_k(self, x, n=BASE_N):
        return np.stack((x, np.repeat(n, x.shape[0]))).T

    def separate_pdf(self, x):
        x = self.add_n_to_k(x)
        return super().separate_pdf(x)

    def _cdf_separate(self, X):
        """ should output  array of shape: (self.n_components, n_samples"""
        zero_component = (X[:, 0, 0] <= 0).astype(float)[..., np.newaxis]
        if len(X.shape) == 1:
            X = X[np.newaxis]
        non_zero = scipy.stats.binom.cdf(X[:, 0] + self.eps, X[:,1], **self.params)
        return np.concatenate((zero_component, non_zero), axis=-1)

    def _sf_separate(self, X):
        zero_component = (X[:, 0, 0] >= 0).astype(float)[..., np.newaxis]
        if len(X.shape) == 1:
            X = X[np.newaxis]
        non_zero = scipy.stats.binom.sf(X[:, 0] - self.eps, X[:,1], **self.params)
        return np.concatenate((zero_component, non_zero), axis=-1)



# Generic Dataset based errors

# not implemented


# API User Interface Error ################################

class UserInputError(Exception):
    status_code = 400
    severity = 'low'

    def __init__(self, message, **kwargs):
        super().__init__(message)
        self.message = message
        self.details = kwargs


    def to_dict(self):
        return {'message': self.message,
                'type': "InputError",
                'severity': self.severity,
                'details': self.details}

class DataError(UserInputError):
    def __init__(self, message, **kwargs):
        super().__init__(message, **kwargs)

# Train Time Errors ###########################################################

class ModelNotFit(Exception):
    status_code = 500

    def __init__(self, reason, model_type):
        message = model_type + ' ' + reason
        super().__init__(message)
        self.message = message


class ModelFailedToFit(ModelNotFit):
    def __init__(self, model_type):
        super().__init__(reason='UnkownError', model_type=model_type)


class SingularMatrix(ModelNotFit):
    def __init__(self, model_type):
        super().__init__(reason='SingluarMatrix', model_type=model_type)


class SampleSizeTooSmall(ModelNotFit):
    def __init__(self, model_type):
        super().__init__(reason='SampleSizeTooSmall', model_type=model_type)




class NotEnoughVariation(ModelNotFit):
    def __init__(self, model_type):
        super().__init__(reason='NotEnoughVariation', model_type=model_type)


### Specific Likelihood Errors

class LikelihoodError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(message)


class LikelihoodDecreasingError(LikelihoodError):
    def __init__(self, message):
        super().__init__('Likelihood is Decreasing: '+ message)


class LikelihoodNaNError(LikelihoodError):
    def __init__(self):
        super().__init__('Error: NaN in likelihood')


class AllZeros(ModelNotFit):
    def __init__(self, model_type):
        super().__init__(reason='AllDataZero', model_type=model_type)


import numpy as np
import scipy.stats
from typing import Tuple
from ..base import MixtureModel
from ...utils.static import Params, eps


class PoissonMixtureModel(MixtureModel):
    _param_signature_ = {'mu': {'type': float, 'max': np.inf, 'min': 0},}
    _support = {'min': 0-eps, 'max': np.inf, 'dtype': int}
    pdf_thresholder = 10e-8

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = 'PoissonMixtureModel'
        self.discrete = True
        self.n_variables = 1

    def likelihood(self, weights, data, log=False, **params):
        likelihoods = scipy.stats.poisson.pmf(data, **params)
        likelihood_weighted = likelihoods * weights
        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def _fit_deterministic_(self, X):
        rate = X.mean()
        weights = np.array([1])
        return {'mu': rate.reshape(1)}, weights

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        weights = np.ones(n_components) / n_components

        unique, counts = np.unique(X, return_counts=True)
        try:
            rates = np.random.choice(unique, n_components, replace=False, p=counts / counts.sum()).astype(np.float32)
        except ValueError as not_enough:
            rates = X.mean() * (np.random.random(n_components) / (n_components) + .5)

        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=1)
        log_likelihoods = [-np.inf]

        for iteration in range(self.max_iterations):
            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False, mu=rates)
            r = (eps + likelihood_weighted.T) / (likelihood_weighted + self.eps).sum(axis=1)  # todo fix

            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)

            if self.check_convergence(log_likelihoods):
                break
            # Maximization
            rates = (r * X).sum(axis=1) / r.sum(axis=1)
            weights = r.sum(axis=1) / r.sum()
            if rates.sum() == 0:  # cant both be zero
                rates[np.argmin(weights)] = .01

        return {'mu': rates}, weights, log_likelihoods[-1]


    def _generate_component(self, component, n):
        return scipy.stats.poisson.rvs(mu=self.params['mu'][component], size=n)

    def _cdf_separate(self, X):
        return scipy.stats.poisson.cdf(X + self.eps, **self.params)

    def _sf_separate(self, X):
        return scipy.stats.poisson.sf(X - self.eps, **self.params)


import numpy as np
import scipy.stats
from typing import Tuple, Union
from ..base import MixtureModel
from ...utils.static import Params, eps

class LogNormalMixtureModel(MixtureModel):
    """
        This class is the implementation of the Gaussian Mixture Models
        inspired by sci-kit learn implementation.
    """
    _param_signature_ = {'mu': {'type': float, 'max': np.inf, 'min': -np.inf},
                         'sigma': {'type': float, 'max': np.inf, 'min': 0}}
    _support = {'min':0-eps, 'max':np.inf, 'dtype': float}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.n_variables = 1
        self.name = 'LogNormalMixtureModel'
        self.zero_truncated = True

    def _fit_deterministic_(self, X) -> Tuple[Params, np.ndarray]:
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.lognorm.html
        s, _, scale = scipy.stats.lognorm.fit(X, floc=0)
        sigma = s
        mu = np.log(scale)
        return {'mu': mu.reshape(1), 'sigma': sigma.reshape(1)}, np.array([1])

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        weights = np.ones(n_components) / n_components
        log_X = np.log(X)
        mu = log_X[np.random.choice(range(X.shape[0]), n_components)]
        sigma = np.random.random(n_components) * (log_X.var() / n_components)

        log_X_rep = np.repeat(log_X[..., np.newaxis], n_components, axis=1)
        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=1)
        log_likelihoods = [-np.inf]
        for iteration in range(self.max_iterations):
            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False, mu=mu, sigma=sigma)
            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)

            if self.check_convergence(log_likelihoods):
                break

            r = likelihood_weighted / (likelihood_weighted + self.eps).sum(axis=1)[..., np.newaxis]
            # maximization

            mu = np.average(log_X_rep, axis=0, weights=r)
            sigma = np.sqrt(np.average(np.square((log_X_rep - mu)), weights=r, axis=0))
            weights = r.sum(axis=0) / r.sum()

        return {'mu': mu, 'sigma': sigma}, weights, log_likelihoods[-1]

    def likelihood(self, weights: np.ndarray, data: np.ndarray, mu, sigma, log=False
                   ) -> Union[np.ndarray, Union[float, np.float64]]:
        likelihoods = scipy.stats.lognorm.pdf(data, scale=np.exp(mu), s=sigma, loc=0)
        likelihood_weighted = likelihoods * weights
        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted


    def _generate_component(self, component, n):
        return scipy.stats.lognorm.rvs(
            scale=np.exp(self.params['mu'][component]), s=self.params['sigma'][component], size=n)
    def _cdf_separate(self, X):
        return scipy.stats.lognorm.cdf(X, scale=np.exp(self.params['mu']), s=self.params['sigma'], loc=0)

    def _sf_separate(self, X):
        return scipy.stats.lognorm.sf(X, scale=np.exp(self.params['mu']), s=self.params['sigma'], loc=0)


import numpy as np
import json
from typing import Union, Dict
from .models import model_dict
from ..utils.static import DATETIME_FORMAT
import datetime as dt


def load_model_from_file(model: Union[str, Dict]):
    if isinstance(model, str):
        if 'json' == model[-4:].lower():
            with open(model, 'r', encoding='utf8') as fo:
                model = fo.read()
        return load_model_from_config(json.loads(model))
    elif isinstance(model, dict):
        return load_model_from_config(model)
    raise NotImplementedError




def load_model_from_config(config: Dict):
    model = model_dict[config['type']]()
    params = config['params']
    for key in params.copy():
        param = np.array(params[key])
        if len(param.shape) == 0:
            param = param.reshape(-1)
        params[key] = param
    model.params = params
    model.weights = np.array(config['weights'])
    model.n_components = config['n_components']
    model.max_iterations = config['max_iterations']
    model.n_variables = config['n_variables']
    model.sample_size = config['sample_size']
    model.datetime_fit = dt.datetime.strptime(config['datetime_fit'],DATETIME_FORMAT
                                              ) if 'datetime_fit' in config else None
    model.seconds_to_fit = config.get('seconds_to_fit', None)
    model.isfit = True
    model.converged = config.get('converged', True)
    return model





import numpy as np
import scipy.stats
from typing import Tuple, Union
from ..base import MixtureModel
from ...utils.static import Params, eps
from sklearn.mixture import GaussianMixture
import warnings


class NormalMixtureModel(MixtureModel):
    """
        This class is the implementation of the Gaussian Mixture Models
        inspired by sci-kit learn implementation.
    """
    _param_signature_ = {'loc': {'type': float, 'max': np.inf, 'min': -np.inf},
                         'scale': {'type': float, 'max': np.inf, 'min': 0}, }
    _support = {'min': -np.inf, 'max': np.inf, 'dtype': float}

    def __init__(self, tied_variance=False, zero_truncated=False, **kwargs):
        super().__init__(zero_truncated=zero_truncated, **kwargs)
        self.n_variables = 1
        self.name = 'NormalMixtureModel'
        self.tied_variance = tied_variance

    def _fit_deterministic_(self, X):
        loc = X.mean()
        scale = X.std()
        weights = np.array([1])
        return {'loc': loc.reshape(1), 'scale': scale.reshape(1)}, weights

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        weights = np.ones(n_components) / n_components
        means = X[np.random.choice(range(X.shape[0]), n_components)]
        variances = np.random.random(n_components) * (X.var() / n_components)

        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=1)
        log_likelihoods = [-np.inf]
        for iteration in range(self.max_iterations):
            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False, loc=means,
                                                  scale=np.sqrt(variances))
            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)

            if self.check_convergence(log_likelihoods):
                break

            r = (eps + likelihood_weighted.T) / (likelihood_weighted + eps).sum(axis=1)

            # maximization
            means = (r * X).sum(axis=1) / (r + self.eps).sum(axis=1)
            variances = (r * np.square(X_rep - means + self.eps ** 10).T).sum(axis=1) / (r + self.eps).sum(axis=1)
            if self.tied_variance:
                variances = np.repeat(variances.mean(), n_components)
            weights = r.mean(axis=1)

        return {'loc': means, 'scale': np.sqrt(variances)}, weights, log_likelihoods[-1]

    def likelihood(self, weights: np.ndarray, data: np.ndarray, log=False, **params
                   ) -> Union[np.ndarray, Union[float, np.float64]]:
        likelihoods = scipy.stats.norm.pdf(data, **params)
        likelihood_weighted = likelihoods * weights
        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def _cdf_separate(self, X):
        return scipy.stats.norm.cdf(X, **self.params)

    def _sf_separate(self, X):
        return scipy.stats.norm.sf(X, **self.params)

    def _generate_component(self, component, n):
        return scipy.stats.norm.rvs(loc=self.params['loc'][component], scale=self.params['scale'][component], size=n)


class NormalLocationMixtureModel(NormalMixtureModel):
    """
        This class is the implementation of the Gaussian Mixture Models
        inspired by sci-kit learn implementation.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs, tied_variance=True)
        self.name = 'NormalLocationMixtureModel'


class NormalMixtureModelSklearn(NormalMixtureModel):
    """
        This class is the implementation of the Gaussian Mixture Models
        wrapping the sci-kit learn implementation.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = 'NormalMixtureModelSklearn'
        self.skip_deterministic = True
        self.n_iter = 1

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        if len(X.shape) == 1:
            X = X[..., np.newaxis]
        model = GaussianMixture(n_components)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            model.fit(X)
        loc = model.means_.reshape(-1)
        scale = np.sqrt(model.covariances_.reshape(-1))
        weights = model.weights_
        X_rep = np.repeat(X, n_components, axis=1)
        log_likelihood = self.likelihood(weights=weights, data=X_rep, log=True, loc=loc, scale=scale)
        return {'loc': loc, 'scale': scale}, weights, log_likelihood



import numpy as np
import scipy.stats
from typing import Tuple, Union
from ..base import MixtureModel
from ...utils.exceptions import ModelFailedToFit
from ...utils.static import Params, eps


class BetaMixtureModel(MixtureModel):
    _param_signature_ = {'a': {'type': float, 'max': np.inf, 'min': 0},
                         'b': {'type': float, 'max': np.inf, 'min': 0},}
    _support = {'min':0-eps, 'max':1+eps, 'dtype': float}

    def __init__(self, single_component_bimodal=False, **kwargs):
        super().__init__(**kwargs)
        self.name = 'BetaMixtureModel'
        self.zero_truncated = True
        self.n_variables = 1
        self.min_components = 1
        self.single_component_bimodal = single_component_bimodal
        self.eps = 1e-8


    def _fit_deterministic_(self, X: np.ndarray) -> Tuple[Params, np.ndarray]:
        try:
            a, b, loc, scale = scipy.stats.beta.fit(X, floc=0, fscale=1)
            if ((a < 1) & (b < 1)) and (self.single_component_bimodal is False): # is this good or bad?
                raise RuntimeError('a and b cannot both be less than 1')
        except RuntimeError as e:
            raise ModelFailedToFit(self.name)
        return {'a': a[..., np.newaxis], 'b': b[..., np.newaxis]}, np.array([1])

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        weights = np.ones(n_components) / n_components
        alphas, betas = 1 + np.random.random(size=n_components), 1 + np.random.random(size=n_components)
        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=-1)
        log_likelihoods = [-np.inf]
        for iteration in range(self.max_iterations):
            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False, a=alphas, b=betas)
            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)

            if self.check_convergence(log_likelihoods):
                break

            r = (eps + likelihood_weighted) / (likelihood_weighted + eps).sum(axis=1)[..., np.newaxis]

            # maximization
            gmeans = scipy.stats.gmean(X_rep, weights=r, axis=0)
            gmeans_1minus = scipy.stats.gmean(1 - X_rep, weights=r, axis=0)
            alphas = .5 + gmeans / (2*(1-gmeans - gmeans_1minus))
            betas = .5 + gmeans_1minus / (2 * (1 - gmeans - gmeans_1minus))
            weights = r.mean(axis=0)

        return {'a': alphas, 'b':betas}, weights, log_likelihoods[-1]

    def likelihood(self, weights: np.ndarray, data: np.ndarray, log=False, **params) -> Union[
        np.ndarray, Union[float, np.float64]]:
        likelihoods = scipy.stats.beta.pdf(data, **params, loc=0, scale=1)
        likelihood_weighted = likelihoods * weights
        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def _generate_component(self, component, n):
        return scipy.stats.beta.rvs(a=self.params['a'][component], b=self.params['b'][component], size=n)

    def _cdf_separate(self, X):
        return scipy.stats.beta.cdf(X, **self.params)

    def _sf_separate(self, X):
        return scipy.stats.beta.sf(X, **self.params)


import numpy as np
import scipy.stats
from typing import Tuple, Union
from ..base import MixtureModel
from ...utils.static import  Params, eps


class ZeroInflatedExponentialMixtureModel(MixtureModel):
    _param_signature_ = {'scale': {'type': float, 'max': np.inf, 'min': 0}}
    _support = {'min': 0 - eps, 'max': np.inf, 'dtype': float}
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "ZeroInflatedExponentialMixtureModel"
        self.zero_inflated = True
        self.min_components = 2
        self.n_variables = 1

    def likelihood(self, weights: np.ndarray, data: np.ndarray, log=False, **params) -> Union[
        np.ndarray, Union[float, np.float64]]:
        # **params are scale = 1 / lambda
        likelihood_zero = (data[:, 0] == 0).astype(np.int64)[:, np.newaxis]
        likelihood = scipy.stats.expon.pdf(data[:, 1:], **params)
        likelihood[data[:, 0] == 0] = 0  # put them to zero

        likelihood_weighted = np.concatenate((likelihood_zero, likelihood), axis=1) * weights

        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def _fit_deterministic_(self, X: np.ndarray) -> Tuple[Params, np.ndarray]:
        # https://en.wikipedia.org/wiki/Exponential_distribution#Parameter_estimation
        params, weights, ll = self._fit_(X, self.min_components)
        return params, weights

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        pi = (X == 0).mean()
        weights = np.hstack((pi, (1 - pi) * np.ones(n_components - 1) / (n_components - 1)))

        lambdas = 1 / (X.mean() * (np.random.random(n_components - 1) + .5))
        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=1)
        log_likelihoods = [-np.inf]

        for iteration in range(self.max_iterations):
            # calculate the maximum likelihood of each observation xi

            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False, scale=1 / lambdas)
            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)

            if self.check_convergence(log_likelihoods):
                break
            r = (eps + likelihood_weighted) / (likelihood_weighted + self.eps).sum(axis=1)[..., np.newaxis]
            # Maximization
            lambdas = 1 / np.average(X_rep[:, 1:], weights=r[:, 1:], axis=0)
            weights = r.sum(axis=0) / r.sum()

        lambdas[lambdas == np.inf] = 1 / self.eps  # hotfix
        return {'scale': 1 / lambdas}, weights, log_likelihoods[-1]

    def _cdf_separate(self, X):
        zero_component = (X[:, 0] <= 0).astype(float)[..., np.newaxis]
        non_zero = scipy.stats.expon.cdf(X, **self.params)
        return np.concatenate((zero_component, non_zero), axis=-1)

    def _sf_separate(self, X):
        zero_component = (X[:, 0] >= 0).astype(float)[..., np.newaxis]
        non_zero = scipy.stats.expon.sf(X, **self.params)
        return np.concatenate((zero_component, non_zero), axis=-1)

    def _generate_component(self, component, n):
        return scipy.stats.expon.rvs(scale=self.params['scale'][component-1], size=n)



import numpy as np
import scipy.stats
from typing import Tuple, Union, Callable
from ..base import MixtureModel
from ...utils.static import Params, eps

BASE_N = 100


class BinomialMixtureModel(MixtureModel):
    _param_signature_ = {'p': {'type': float, 'max': 1 + eps, 'min': 0},}
    _support = {'min': 0, 'max': np.inf, 'dtype':int}
    def __init__(self, standardize=False, **kwargs):
        super().__init__(**kwargs)
        self.name = 'BinomialMixtureModel'
        self.discrete = True
        self.n_variables = 2
        self.standardize = standardize

    def _fit_deterministic_(self, X: np.ndarray) -> Tuple[Params, np.ndarray]:
        assert len(X.shape) == 2
        assert X.shape[1] == 2
        ps = (X[:, 0] / X[:, 1])
        p = ps.mean()[..., np.newaxis]
        weights = np.array([1])
        return {'p': p}, weights

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        weights = np.ones(n_components) / n_components
        ps = (X[:, 0].sum() / X[:, 1].sum()) * np.random.uniform(.5, 1.5, size=n_components)
        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=-1)
        log_likelihoods = [-np.inf]
        for iteration in range(self.max_iterations):
            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False, p=ps)
            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)

            if self.check_convergence(log_likelihoods):
                break

            r = (eps + likelihood_weighted) / (likelihood_weighted + eps).sum(axis=1)[..., np.newaxis]

            # maximization
            ps = np.average(X_rep[:, 0, :] / X_rep[:, 1, :], axis=0, weights=r)
            weights = r.mean(axis=0)

        return {'p': ps}, weights, log_likelihoods[-1]

    def likelihood(self, weights: np.ndarray, data: np.ndarray, log=False, **params) -> Union[
        np.ndarray, Union[float, np.float64]]:
        likelihoods = scipy.stats.binom.pmf(n=data[:, 1, :], k=data[:, 0, :], **params)
        likelihood_weighted = likelihoods * weights
        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def _generate_component(self, component, n: int, k: Union[Callable, int, np.ndarray]):
        if isinstance(k, int):
            k = np.array(k)
        elif callable(k):
            k = k(n)

        if len(k.shape) > 1:
            if k.shape[1] == 1:
                k = k.flatten()
            else:
                raise Exception(f'Dont know how to handle k of shape {k.shape}')
        elif len(k.shape) < 1:
            k = k[np.newaxis]

        if k.shape[0] == self.n_components:
            k = k[component]

        if k.shape == (n,):
            pass
        elif k.shape in [(1,), ()]:
            k = np.repeat(k, repeats=n)
        else:
            raise Exception(f'Dont know how to handle k of shape: {k.shape}')

        return np.stack((scipy.stats.binom.rvs(n=k, p=self.params['p'][component], size=n), k)).T

    def separate_pdf(self, x):
        x = np.stack((x, np.repeat(BASE_N, x.shape[0]))).T
        return super().separate_pdf(x)

    def _cdf_separate(self, X):
        if len(X.shape) == 1:
            X = X[np.newaxis]
        return scipy.stats.binom.cdf(X[:,0] + self.eps, X[:,1], **self.params)

    def _sf_separate(self, X):
        if len(X.shape) == 1:
            X = X[np.newaxis]
        return scipy.stats.binom.sf(X[:,0] - self.eps, X[:,1], **self.params)


from typing import Dict, Union, List
from numpy import ndarray

COMPONENT_MIN = 30
DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S.%f'
eps = 1e-128
MIN_ITERATIONS = 25

uni_params = Dict[str, Union[List, ndarray, float]]
Params = Union[uni_params, Dict[str, uni_params]]


class DISTRIBUTIONS:
    NORMAL: str = 'NORMAL'
    LOGNORMAL: str = 'LOGNORMAL'
    ZINORMAL: str = 'ZEROINFLATEDNORMAL'
    POISSON: str = 'POISSON'
    ZIPOISSON: str = 'ZEROINFLATEDPOISSON'
    GEOMETRIC: str = 'GEOMETRIC'
    ZETA: str = 'ZETA'
    BINOMIAL: str = 'BINOMIAL'
    EXPONENTIAL: str = "EXPONENTIAL"
    ZIEXPONENTIAL: str = 'ZEROINFLATEDEXPONENTIAL'
    STUDENTST: str = 'STUDENTST'
    CAUCHY: str = 'CAUCHY'
    BETA: str = 'BETA'
    DISCRETE: list = [POISSON, ZIPOISSON,  GEOMETRIC, ZETA]






from .BetaMixtureModel import BetaMixtureModel
from .BinomialMixtureModel import BinomialMixtureModel
from .CauchyMixtureModel import CauchyMixtureModel
from .ExponentialMixtureModel import ExponentialMixtureModel
from .GeometricMixtureModel import GeometricMixtureModel
from .LogNormalMixtureModel import LogNormalMixtureModel
from .NormalMixtureModel import NormalMixtureModel, NormalLocationMixtureModel, NormalMixtureModelSklearn
from .PoissonMixtureModel import PoissonMixtureModel
from .StudentsTMixtureModel import StudentsTMixtureModel
from .ZeroInflatedBinomialMixtureModel import ZeroInflatedBinomialMixtureModel
from .ZeroInflatedExponentialMixtureModel import ZeroInflatedExponentialMixtureModel
from .ZeroInflatedNormalMixtureModel import ZeroInflatedNormalMixtureModel
from .ZeroInflatedPoissonMixtureModel import ZeroInflatedPoissonMixtureModel
from .ZetaMixtureModel import ZetaMixtureModel



import numpy as np
import scipy.stats
from typing import Tuple, Union
from ..base import MixtureModel
from ...utils.static import Params, eps
from sklearn.mixture import GaussianMixture
import warnings


class ZeroInflatedNormalMixtureModel(MixtureModel):
    """
        This class is the implementation of the Gaussian Mixture Models
        inspired by sci-kit learn implementation.
    """
    _param_signature_ = {'loc': {'type': float, 'max': np.inf, 'min': -np.inf},
                         'scale': {'type': float, 'max': np.inf, 'min': 0}, }
    _support = {'min': -np.inf, 'max': np.inf, 'dtype': float}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.n_variables = 1
        self.min_components = 2
        self.zero_inflated = True
        self.name = 'ZeroInflatedNormalMixtureModel'

    def _fit_deterministic_(self, X):
        pi = (X == 0).mean()
        loc = X[X != 0].mean()
        scale = X[X != 0].std()
        weights = np.hstack((pi, (1 - pi)))
        return {'loc': loc.reshape(1), 'scale': scale.reshape(1)}, weights

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        pi = (X == 0).mean()
        weights = np.hstack((pi, (1 - pi) * np.ones(n_components - 1) / (n_components - 1)))
        loc = X[X != 0].mean() * (np.random.random(n_components - 1) / (n_components - 1) + .5)
        scale = X[X != 0].std() * (np.random.random(n_components - 1) / (n_components - 1) + .5)

        log_likelihoods = [-np.inf]
        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=1)
        for iteration in range(self.max_iterations):
            # calculate the maximum likelihood of each observation xi

            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False,
                                                  loc=loc, scale=scale)
            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)

            if self.check_convergence(log_likelihoods):
                break
            r = (eps + likelihood_weighted) / (likelihood_weighted + self.eps).sum(axis=1)[..., np.newaxis]

            # maximization
            loc = (r[:, 1:] * X_rep[:, 1:]).sum(axis=0) / (r[:, 1:] + self.eps).sum(axis=0)
            scale = np.sqrt(
                (r[:, 1:] * np.square(X_rep[:, 1:] - loc + self.eps ** 10)).sum(axis=0) / (r[:, 1:] + self.eps).sum(
                    axis=0))
            weights = r.sum(axis=0) / r.sum()

        return {'loc': loc, 'scale': scale}, weights, log_likelihoods[-1]

    def likelihood(self, weights: np.ndarray, data: np.ndarray,
                   log=False, **params) -> Union[np.ndarray, Union[float, np.float64]]:

        likelihood_normal = scipy.stats.norm.pdf(x=data[:, 1:], **params)
        likelihood_zero = (data[:, 0] == 0).astype(np.int64)[:, np.newaxis]
        likelihood_weighted = np.concatenate((likelihood_zero, likelihood_normal), axis=1) * weights

        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def _generate_component(self, component, n):
        return scipy.stats.norm.rvs(loc=self.params['loc'][component-1], scale=self.params['scale'][component-1], size=n)

    def _cdf_separate(self, X):
        non_zero = scipy.stats.norm.cdf(X, **self.params)
        zero_component = (X[:, 0] <= 0).astype(float)[..., np.newaxis]
        return np.concatenate((zero_component, non_zero), axis=-1)

    def _sf_separate(self, X):
        non_zero = scipy.stats.norm.sf(X, **self.params)
        zero_component = (X[:, 0] >= 0).astype(float)[..., np.newaxis]
        return np.concatenate((zero_component, non_zero), axis=1)


class ZeroInflatedNormalMixtureModelSklearn(ZeroInflatedNormalMixtureModel):
    """
        This class is the implementation of the Gaussian Mixture Models
        wrapping the sci-kit learn implementation.
    """

    def __init__(self, max_iterations=2000):
        # todo
        raise NotImplementedError
        super().__init__(max_iterations)
        self.name = 'ZeroInflatedNormalMixtureModelSklearn'
        self.skip_deterministic = True

    def _fit_deterministic(self, X):
        return self._fit_(X, self.min_components)

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        if len(X.shape) == 1:
            X = X[..., np.newaxis]
        pi = (X == 0).mean()
        X_nz = X[X != 0]

        model = GaussianMixture(n_components - 1)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            model.fit(X_nz)
        loc = model.means_.reshape(-1)
        scale = np.sqrt(model.covariances_.reshape(-1))
        weights = np.hstack((pi, model.weights_))
        X_rep = np.repeat(X, n_components, axis=1)
        log_likelihood = self.likelihood(weights=weights, data=X_rep, log=True, loc=loc, scale=scale)
        return {'loc': loc, 'scale': scale}, weights, log_likelihood

    def _cdf_separate(self, X):

        return scipy.stats.norm.cdf(X, **self.params)

    def _sf_separate(self, X):
        return scipy.stats.norm.sf(X, **self.params)


from .univariate import (
    BetaMixtureModel,
    BinomialMixtureModel,
    CauchyMixtureModel,
    ExponentialMixtureModel,
    GeometricMixtureModel,
    LogNormalMixtureModel,
    NormalMixtureModel,
    NormalLocationMixtureModel,
    NormalMixtureModelSklearn,
    PoissonMixtureModel,
    StudentsTMixtureModel,
    ZeroInflatedBinomialMixtureModel,
    ZeroInflatedExponentialMixtureModel,
    ZeroInflatedNormalMixtureModel,
    ZeroInflatedPoissonMixtureModel,
    ZetaMixtureModel
)

from ..utils.static import DISTRIBUTIONS

model_dict = {
    BetaMixtureModel.__name__: BetaMixtureModel,
    BinomialMixtureModel.__name__: BinomialMixtureModel,
    CauchyMixtureModel.__name__: CauchyMixtureModel,
    ExponentialMixtureModel.__name__: ExponentialMixtureModel,
    GeometricMixtureModel.__name__: GeometricMixtureModel,
    LogNormalMixtureModel.__name__: LogNormalMixtureModel,
    NormalMixtureModel.__name__: NormalMixtureModel,
    NormalMixtureModelSklearn.__name__: NormalMixtureModelSklearn,
    NormalLocationMixtureModel.__name__:NormalLocationMixtureModel,
    PoissonMixtureModel.__name__: PoissonMixtureModel,
    StudentsTMixtureModel.__name__: StudentsTMixtureModel,
    ZeroInflatedBinomialMixtureModel.__name__: ZeroInflatedBinomialMixtureModel,
    ZeroInflatedExponentialMixtureModel.__name__: ZeroInflatedExponentialMixtureModel,
    ZeroInflatedNormalMixtureModel.__name__: ZeroInflatedNormalMixtureModel,
    ZeroInflatedPoissonMixtureModel.__name__: ZeroInflatedPoissonMixtureModel,
    ZetaMixtureModel.__name__: ZetaMixtureModel
}


UNIVARIATE = [BetaMixtureModel, BinomialMixtureModel, CauchyMixtureModel, ExponentialMixtureModel,
              GeometricMixtureModel, LogNormalMixtureModel, NormalMixtureModel, PoissonMixtureModel,
              StudentsTMixtureModel, ZeroInflatedBinomialMixtureModel, ZeroInflatedExponentialMixtureModel,
              ZeroInflatedNormalMixtureModel, ZeroInflatedPoissonMixtureModel, ZetaMixtureModel]


IMPLEMENTED = UNIVARIATE

ExternalModelNameDict = {DISTRIBUTIONS.NORMAL: NormalMixtureModelSklearn,
                         DISTRIBUTIONS.LOGNORMAL: NormalMixtureModelSklearn,
                         DISTRIBUTIONS.POISSON: PoissonMixtureModel,
                         DISTRIBUTIONS.STUDENTST: StudentsTMixtureModel,
                         DISTRIBUTIONS.ZIPOISSON: ZeroInflatedPoissonMixtureModel,
                         DISTRIBUTIONS.ZINORMAL: ZeroInflatedNormalMixtureModel,
                         DISTRIBUTIONS.EXPONENTIAL: ExponentialMixtureModel,
                         DISTRIBUTIONS.ZIEXPONENTIAL: ZeroInflatedExponentialMixtureModel,
                         DISTRIBUTIONS.CAUCHY: CauchyMixtureModel,
                         DISTRIBUTIONS.BINOMIAL: BinomialMixtureModel,
                         DISTRIBUTIONS.GEOMETRIC: GeometricMixtureModel,
                         DISTRIBUTIONS.BETA: BetaMixtureModel,
                         }



import setuptools
from src import __version__, __author__, __author_email__

with open("readme.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as fo:
    install_requires = fo.read().split('\n')

setuptools.setup(
    name="anomix",
    version=__version__,
    author=__author__,
    author_email=__author_email__,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
    ],
    description="Mixture Models for anomaly detection",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TRSS-Research/anomix",
    project_urls={},
    entry_points={},
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src") ,
    install_requires=install_requires,
    python_requires=">=3.6",
)



import abc
import datetime as dt
import json
import logging
from typing import Optional, Tuple, Union, List, Dict, Iterable
import numpy as np
import matplotlib.pyplot as plt
from ..utils.exceptions import (
    SampleSizeTooSmall,
    AllZeros,
    LikelihoodNaNError,
    LikelihoodError,
    LikelihoodDecreasingError,
    ModelFailedToFit,
    UserInputError,
    DataError
)

from ..utils.static import Params, COMPONENT_MIN, MIN_ITERATIONS, DATETIME_FORMAT
from ..utils.utils import simplify_params, sort_params,checkfit

logger = logging.getLogger(__name__)
BASE_PLOT_MAX = 1000

class PresetParamsError(UserInputError):
    pass


class ParamSizeMismatchError(PresetParamsError):
    pass


class ParamTypeError(PresetParamsError):
    pass


class ParamBoundsError(PresetParamsError):
    pass


class IdenticalComponentError(PresetParamsError):
    pass


class MissingKeyError(UserInputError):
    pass


class MixtureModel(object):
    """
        Abstract base class for Mixture Models, to be subclassed with models designed for specific distributions.
         This base class contains generic high level functions which call the subclass specific functions (primarily
         `likelihood` and `_fit_`. `likelihood` reflects the mixture model likelihood for the data, while `_fit_`
         implements the specific distributions flavor of the EM algorithm.
            E-M algorithm overview:  https://en.wikipedia.org/wiki/Expectation%E2%80%93maximization_algorithm
    """
    _param_signature_ = {'None': {'type': type(None), 'max': np.inf, 'min': -np.inf}}
    _support = {'min':-np.inf, 'max':np.inf, 'dtype':float}
    pdf_thresholder = 10e-5
    def __init__(self, **kwargs) -> None:
        """
            Initialize the model, set some basic class variables
        """
        self.eps = 1e-8
        self.tol = 1e-2
        self.n_iter = 5
        self.name = 'MixtureModel'


        # general model train configurations
        self.max_iterations = kwargs.get('max_iterations', 100)
        self.min_components = kwargs.get('min_components', 1)
        self.max_components = kwargs.get('max_components', 3)
        self.unimodal = kwargs.get('unimodal', False)
        self.sort_params = kwargs.get('sort_params', False)


        # default Values to be overwritten in subclasses
        self.skip_deterministic = False
        self.zero_truncated = False
        self.discrete = False
        self.zero_inflated = False

        # model parameters to be edited by train process
        self.params = dict()
        self.n_components = None
        self.weights = None
        self.sample_size = None
        self.n_variables = None
        self.converged = False
        self.isfit = False

        self.datetime_fit = None
        self.seconds_to_fit = 0

    def subset_zeros(self, X):
        """
            remove zeros from the data. Useful when zeros are representative of nullity.
             If multivariate, then remove if the model is 'zero truncated' then we drop any instance we observe a zero
        :param X: data to censor
        :return: censored data
        """
        if isinstance(self.zero_truncated, bool):
            if self.zero_truncated:  # if our model cannot handle zeros we will drop all observations of zero.
                if len(X.shape) == 1:
                    X = X[X != 0]
                else:
                    X = X[(X != 0).all(axis=1)]
        elif isinstance(self.zero_truncated, tuple):
            for i, truncated in enumerate(self.zero_truncated):
                if truncated:
                    X = X[X[:, i] != 0, :]
        return X

    def __repr__(self):
        return self.__str__()

    def fit(self, X: np.ndarray, overwrite=False) -> None:
        """
            Fit an mixture model of arbitrary number of components, and distributions.
        :param X: data to be fit.
        :param overwrite: bool, will not overwrite model if set to False
        :return: None
        """
        if self.isfit and not overwrite:
            raise ModelFailedToFit('Model already fit. If you want to overwrite, set overwrite=True)')

        self.datetime_fit = dt.datetime.now()
        logger.debug(f"Begin Fitting {self.name}")
        X = self.subset_zeros(X)

        if len(X.shape) == 1:  # set the number of variables and sample size for easy future reference.
            self.sample_size, self.n_variables = X.shape[0], 1
            if (X == 0).all():
                raise AllZeros(self.name)
        else:
            self.sample_size, self.n_variables = X.shape[0], X.shape[1]
            if self.n_variables > 1:
                if (X == 0).all(axis=1).any():
                    raise AllZeros(self.name)
            else:
                X = X.reshape(-1)

        self.check_data_supported(X)

        if isinstance(self.discrete, bool):
            if self.discrete:
                X = np.ceil(X).astype(np.int64)
        elif isinstance(self.discrete, tuple):
            for v in range(self.n_variables):
                if self.discrete[v]:
                    X[:, v] = np.ceil(X[:, v]).astype(np.int64)

        max_components = min(self.sample_size // COMPONENT_MIN, self.max_components)
        if max_components < self.min_components:
            # Too few samples to fit a model of any size
            raise SampleSizeTooSmall(self.name)

        elif max_components == self.min_components:
            # Only enough data to fit a unimodal model
            self.n_components = self.min_components

        if (self.n_components is None) and (self.min_components != max_components) and (not self.unimodal):
            # Only choose the number of components if we need to
            logger.debug(f"Finding Optimal Components")
            # choose the optimal number of components
            self.n_components = self.optimize_components(X, max_components)

        # Now fit the actual model
        logger.debug(f"Fitting model with {self.n_components}")
        self.converged = False  # need to reset convergence param
        params, weights, ll = self._fit(X, self.n_components)
        if params is None:
            raise ModelFailedToFit(self.name)
        self.params = params
        self.weights = weights
        self.isfit = True
        
        self.seconds_to_fit = (dt.datetime.now() - self.datetime_fit).seconds + \
                              (dt.datetime.now() - self.datetime_fit).microseconds / 1000000
        logger.debug(f'Completed')

    def preprocess_pred(self, X: Union[int, float, np.ndarray]):
        if not isinstance(X, np.ndarray):
            X = np.array([X])
        self.check_data_supported(X)
        return X

    def check_data_supported(self, X: np.ndarray):
        if len(X.shape) == 1:
            if self.n_variables == 1: pass
            elif self.n_variables != X.shape[0]:
                raise DataError(f'Not sure what to do with data of shape: {X.shape}. Expecting either (1, {self.n_variables}) or ({self.n_variables},)')
        elif X.shape[1] != self.n_variables:
            raise DataError(f'Not sure what to do with data of shape: {X.shape}. Expecting either (1, {self.n_variables}) or ({self.n_variables},)')

        if self.zero_truncated:
            if any(X == 0):
                raise DataError(f'{self.name} is zero truncated, but you provided zeros')
        if (X < self._support['min']).any() or (X > self._support['max']).any():
            raise DataError(f'Outside of Support range. Required: {self._support}, received: {X}')
        if not (X.astype(self._support['dtype']) == X).all():
            raise DataError(f'Cannot cast {X.dtype} to {self._support["dtype"]}')

    @abc.abstractmethod
    def _cdf_separate(self, X):
        raise NotImplementedError

    @abc.abstractmethod
    def _sf_separate(self, X):
        raise NotImplementedError

    def _cdf(self, X):
        return (self._cdf_separate(X) * self.weights) .sum()

    def _sf(self, X):
        return (self._sf_separate(X) * self.weights).sum()

    def cdf_separate(self, X):
        X = np.repeat(X[..., np.newaxis], repeats=self.n_components - self.zero_inflated, axis=-1)
        return self._cdf_separate(X)

    def sf_separate(self, X):
        X = np.repeat(X[..., np.newaxis], repeats=self.n_components - self.zero_inflated, axis=-1)
        return self._sf_separate(X)

    def cdf(self, X):
        X = np.repeat(X[..., np.newaxis], repeats=self.n_components - self.zero_inflated, axis=-1)
        return self._cdf(X)

    def sf(self, X):
        X = np.repeat(X[..., np.newaxis], repeats=self.n_components - self.zero_inflated, axis=-1)
        return self._sf(X)

    def _predict_proba(self, X: np.ndarray, test_lower_tail=True, test_upper_tail=True):
        if test_lower_tail:
            lower_tail = self.cdf_separate(X).reshape((X.shape[0], self.weights.shape[0]))
        else:
            lower_tail = np.ones((X.shape[0], self.weights.shape[0]))

        if test_upper_tail:
            upper_tail = self.sf_separate(X).reshape((X.shape[0], self.weights.shape[0]))
        else:
            upper_tail = np.ones((X.shape[0], self.weights.shape[0]))

        tails = np.stack((lower_tail, upper_tail))
        pvals = np.min(tails, axis=0) * self.weights
        assigned = pvals.argmax(axis=1)
        return pvals[np.ones_like(assigned, dtype=bool), assigned], assigned

    @checkfit
    def predict_proba(self, X, unweighted=False, test_lower_tail=True, test_upper_tail=True):
        X = self.preprocess_pred(X)
        assert test_lower_tail or test_upper_tail, 'Must specify at least one tail to test'
        pval, assigned = self._predict_proba(X, test_lower_tail=test_lower_tail, test_upper_tail=test_upper_tail)
        if unweighted:
            pval /= self.weights[assigned]
        return pval, assigned

    @checkfit
    def predict_anomaly(self, X: [int, float, np.ndarray], threshold: float=.8, unweighted: bool = True,
                        test_lower_tail: bool = True, test_upper_tail: bool = True) -> np.ndarray:
        """
            predict whether input is anomalous in comparison to the fit model, can specify to include the baseline
                component weights as part of the anomaly reporting, or to treat the components as unweighted, full distributions
        :param X: input to evaluate.
        :param threshold:
        :param unweighted:
        :param test_lower_tail: bool, if true tests if value is anomalous in the lower tail
        :param test_upper_tail: bool, if true tests if value is anomalous in the upper tail
        :return:
        """
        if not (test_upper_tail or test_lower_tail):
            raise UserInputError("Cannot predict anomaly without specifying at least one tail")
        X = self.preprocess_pred(X)
        pval, assigned = self.predict_proba(X, unweighted)
        alpha = (1-threshold) / (test_lower_tail + test_upper_tail)
        return pval < alpha

    @checkfit
    def predict(self, X: [int, float, np.ndarray], threshold: float=.8, unweighted: bool = True,
                test_lower_tail: bool = True, test_upper_tail: bool = True) -> np.ndarray:
        return self.predict_anomaly(X=X, threshold=threshold, unweighted=unweighted, test_lower_tail=test_lower_tail,
                                    test_upper_tail=test_upper_tail)

    def check_convergence(self, log_likelihood_list, ignore_assertion=False):
        current, previous = log_likelihood_list[-1], log_likelihood_list[-2]
        if np.isnan(current):
            raise LikelihoodNaNError()
        if len(log_likelihood_list) < MIN_ITERATIONS:
            return False
        if not ignore_assertion:
            if current + self.tol < previous:
                raise LikelihoodDecreasingError(f'{log_likelihood_list[-1]} < {log_likelihood_list[-2]}')
        if current - self.tol < previous:
            self.converged = True
            return True
        else:
            return False

    def get_optimal_loglikelihood(self, X, true_params, true_weights):
        n_components = true_weights.shape[0]
        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=-1)
        return self.likelihood(data=X_rep, weights=true_weights, log=True, **true_params)

    def __str__(self):
        info = {'type': self.name,
                'max_iterations': self.max_iterations,
                'zero_truncated': self.zero_truncated
                }
        if self.isfit:
            params = simplify_params(self.params)
            info.update({'params': params,
                         'weights': self.weights.tolist(),
                         'n_components': int(self.n_components),
                         'n_variables': self.n_variables,
                         'sample_size': self.sample_size,
                         'datetime_fit': dt.datetime.strftime(self.datetime_fit, DATETIME_FORMAT),
                         'seconds_to_fit': self.seconds_to_fit,
                         'converged': self.converged,
                         })

        return json.dumps(info, indent=4, sort_keys=False)

    @checkfit
    def save(self, filepath):
        with open(filepath, 'w', encoding='utf8') as fo:
            fo.write(self.__str__())

    def _fit_deterministic(self, X: np.ndarray) -> Tuple[Union[Params, None], np.ndarray, float]:
        try:
            params, weights = self._fit_deterministic_(X)
        except ModelFailedToFit:
            return None, np.array([1]), -np.inf
        X_rep = np.repeat(X[..., np.newaxis], self.min_components, axis=-1)
        ll = self.likelihood(data=X_rep, weights=weights, log=True, **params)
        self.converged = True
        return params, weights, ll

    @abc.abstractmethod
    def _fit_deterministic_(self, X: np.ndarray) -> Tuple[Params, np.ndarray]:
        raise NotImplementedError

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        raise NotImplementedError

    @staticmethod
    def check_weights(weights):
        if not isinstance(weights, np.ndarray):
            weights = np.array(weights, dtype=float)
        if not np.isclose(weights.sum(), 1).all():
            raise ValueError(f'weights must sum to 1. received: {weights} which sums to {weights.sum()}')
        if (weights < 0).any() or (weights > 1).any():
            raise ValueError(f'weights must all be between 0 and 1. received {weights}')
        return weights

    def preset(self, weights: Union[np.ndarray, Iterable] , overwrite=False, **params):
        if self.isfit:
            if overwrite:
                pass
            else:
                raise UserInputError("Model is already fit! Must specify 'overwrite=True' if you want to reset.")
        weights = self.check_weights(weights)
        self.check_params(weights, params)

        self.weights = weights
        self.params = params
        self.isfit = True
        self.n_components = len(weights)
        self.datetime_fit = dt.datetime.now()

    def check_params(self, weights: np.ndarray, params: dict):
        weights_size = weights.size - int(self.zero_inflated)

        for key in self._param_signature_:  # parameter (e.g loc, scale)
            if not key in params:
                raise MissingKeyError(f'Cannot Find {key}. Expecting Signature like: {self._param_signature_}')
            if not np.array(params[key]).size == weights_size:
                raise ParamSizeMismatchError(f'Number of Component mismatch between weights and {key}. Received weights: {weights}, {key}:{params[key]}')
            for i, arg in enumerate(params[key]):  # which component
                _max, _min = self._param_signature_[key]['max'], self._param_signature_[key]['min']
                _typed = self._param_signature_[key]['type'](arg)
                if not _typed == arg:
                    raise ParamTypeError(
                        f'{key} is of wrong type. Value changed when coerced (coerced to {_typed} from {arg}')
                if (arg >= _max) or (arg <= _min):
                    raise ParamBoundsError(
                        f'{key} is outside of bounds. Received {arg}, but bounded between ({_min:.2f}, {_max:.2f})')
        for i in range(1, weights_size):
            component_param_values = np.array([params[key][i] for key in self._param_signature_])
            other_param_values = np.array([[params[key][j] for key in self._param_signature_] for j in range(weights_size) if j != i ])
            if (component_param_values ==other_param_values).all(axis=1).any():
                raise IdenticalComponentError('Found identical component parameters, cannot be a mixture of identical components')

    def _fit(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        """
            Fit a model multiple times and choose the best option one. This tries to accommodate for dependence on
            initial conditions.
        :param X: data
        :param n_components: number of componets we are fitting this time
        :return: params, weights, minimum log likelihood
        """
        if n_components == self.min_components and not self.skip_deterministic:
            return self._fit_deterministic(X)
        best_params, best_weights = None, None
        max_ll = -np.inf
        for _ in range(self.n_iter):
            # Fit the model several times, choosing the model with the highest log-likelihood
            try:
                params, weights, ll = self._fit_(X, n_components)
            except LikelihoodError as e:
                logger.warning("Likelihood error - " + e.message)
                continue
            if ll > max_ll:
                # if we have a better LL than before, update the 'best'
                best_params, best_weights, max_ll = params, weights, ll
        if self.sort_params:
            best_weights, best_params = sort_params(best_weights, params=best_params, zeros=self.zero_inflated)
        return best_params, best_weights, max_ll

    @abc.abstractmethod
    def likelihood(self, weights: np.ndarray, data: np.ndarray, log=False, **params
                   ) -> Union[np.ndarray, Union[float, np.float64]]:
        """
            likelihood function. this is model and distribution dependent.
        :param weights: generic mixture weights
        :param data: data to fit
        :param log: bool- if true return LogLikelihood
        :param params: kwargs of the distribution to be fit
        :return:
        """
        raise NotImplementedError

    def log_likelihood(self, likelihood_weighted):
        return np.log(likelihood_weighted.sum(axis=1) + self.eps).sum()

    @staticmethod
    def choose_from_loglikelihoods(llikelihoods: np.ndarray, n: int, modifier: float = 2):
        """
            modified BIC (seriously modified)
                This is modified to put more severe penalties on more components
            Chooses the index of the optimal choice criteria. we flipped LL to negative
        :param llikelihoods: array of n_components and associated loglikelihoods
        :param n: number of records
        :param modifier: float modifier to BIC. at 0, no modification.
        :return: index of optimal model
        """
        choice_criteria = - (2 + modifier) * llikelihoods[:, 1] + np.log(n) * (llikelihoods[:, 0] + 1) ** (1 + modifier)
        return np.argmin(choice_criteria)

    def optimize_components(self, X: np.ndarray, max_components) -> int:
        """
            train-test cross-validation to decide the optimal number of components in the model.
                This adds a lot of time to the modeling, we fit the model
                ((max_components-self.min_components) * self.n_iter) times
        :param X: data
        :param max_components: maximum components we are able to fit due to sample sizes
        :return: optimal number of components
        """
        np.random.shuffle(X)
        train = X[:X.shape[0] // 2, ...]
        test = X[X.shape[0] // 2:, ...]
        if (train == 0).all() or (test == 0).all():
            self.n_components = self.min_components
            return self.min_components
        test_llikelihood = []
        for n_components in range(self.min_components, max_components + 1):
            # Find the best model at a given number of components
            params, weights, train_ll = self._fit(train, n_components)
            # potential todo: break if one of the weights is below a certain threshold?
            if params is None:
                test_llikelihood.append((n_components, -np.inf))
                continue
            test_repeat = np.repeat(test[..., np.newaxis], n_components, axis=-1)  # j

            # test this 'best' model on unseen data and get the log-likelihood
            test_ll = self.likelihood(weights=weights, data=test_repeat, log=True, **params)
            test_llikelihood.append((n_components, test_ll))

        test_llikelihood = np.array(test_llikelihood)
        if (test_llikelihood.size == 0) or np.isinf(test_llikelihood[:, 1]).all():  # unable to fit, pass the buck
            logger.warning(f'No Valid Test Likelihoods Passed')
            return self.min_components

        # choose the best number of components based on the log likelihoods
        try:
            optim_n_components = self.choose_from_loglikelihoods(
                test_llikelihood, n=test.shape[0], modifier=2) + self.min_components
        except IndexError:
            logger.warning(f'Choose Component failed: {test_llikelihood}')
            optim_n_components = self.min_components
        self.n_components = optim_n_components
        return optim_n_components

    def generate_zero_inflated_component(self, n):
        component_n = round(self.weights[0] * n)
        return np.zeros(component_n), np.zeros(component_n)

    @checkfit
    def generate(self, n: int, **kwargs):
        """
            generate data given self
        :param n: number of samples
        :return:
        """
        data = []
        labels = []
        if self.zero_inflated:
            zero_data, zero_labels = self.generate_zero_inflated_component(n)
            data.append(zero_data)
            labels.append(zero_labels * 0)
        for i in range(int(self.zero_inflated), self.n_components):
            weight = self.weights[i]
            component_n = round(weight * n)
            data.append(self._generate_component(component=i, n=component_n, **kwargs))
            labels.append(np.ones(component_n) * i)
        data, labels = np.concatenate(data), np.concatenate(labels)
        ix = np.arange(data.shape[0])
        np.random.shuffle(ix)
        return labels[ix], data[ix]

    @abc.abstractmethod
    def _generate_component(self, component: int,  n: int, **kwargs):
        """
            Generate data for a given component index.
            Abstract method implemented by the subclass
            :param component: component index
            :param n: number of samples to generate
        """
        raise NotImplementedError

    def pdf(self, x):
        return self.separate_pdf(x).sum(axis=1)

    def separate_pdf(self, x):
        X = np.repeat(x[..., np.newaxis], repeats=self.n_components, axis=-1)
        return self.likelihood(weights=self.weights, data=X, log=False, **self.params)

    @checkfit
    def get_plotting_range(self):
        """
            identify the minimum and maximum over which we can plot, depending on the 'support' of the distribution
        :return:
        """
        _min = max(self._support['min'], -BASE_PLOT_MAX)
        _max = min(self._support['max'],  BASE_PLOT_MAX)
        X = self.get_plotting_array(_min, _max)
        pdf = self.pdf(X)
        pdf_threshold = pdf.max() * self.pdf_thresholder
        _min = X[np.argmax(pdf > pdf_threshold)]
        _max = X[pdf.shape[0] - np.argmax(pdf[::-1] > pdf_threshold) - 1]
        return _min, _max

    def get_plotting_array(self, _min: float, _max: float) -> np.ndarray:
        """
            get a range of values between _min and _max to plot over
        :return:
        """
        if self.discrete:
            X = np.arange(_min, _max)
        else:
            X = np.linspace(_min, _max, num=10001)
        if self.zero_inflated:
            if not (X == 0).any():
                ix = np.argwhere(X > 0)[0]
                X = np.insert(X, ix, 0)
        return X

    def plot_pdf(self, x_range: Optional[Tuple[float, float]] = None, show=True, smooth_discrete=True,
                 fig_ax: Optional[Tuple] = None):
        """
            Plot the composite likelihood of the mixtures.
            Infers a range if none is provided based on calculated likelihood and support
        :param x_range: minimum and maximum for the x range
        :param show: show figure or not
        :param smooth_discrete: smooth the likelihod for discrete distributions
        :param fig_ax: tuple of figure and axis to plot upon
        :return: (figure, axis)
        """
        if x_range is None:
            x_range = self.get_plotting_range()
        _min, _max = x_range
        X = self.get_plotting_array(_min, _max)
        if fig_ax is None:
            f, ax = plt.subplots(figsize=(12, 8))
        else:
            f, ax = fig_ax

        component_pdfs = self.separate_pdf(X)
        component_list = self.get_components()
        if self.discrete and smooth_discrete:
            from scipy.interpolate import make_interp_spline, BSpline, interp1d

            xnew = np.linspace(X.min(), X.max(), 300)

            component_pdfs = make_interp_spline(X, component_pdfs, k=2)(xnew)  # type: BSpline
            component_pdfs[component_pdfs < 0] = 0
            if self.zero_inflated:
                component_pdfs[np.argwhere(xnew == 0), 0] = self.weights[0]
                component_pdfs[np.argwhere(xnew != 0),0] = 0
            X = xnew
        pdf = component_pdfs.sum(axis=1)

        ax.plot(X, component_pdfs, label=[{key: f'{params[key]:.2f}' for key in params} for params in component_list],
                linewidth=5, alpha=1/(len(component_list) + 1))
        ax.plot(X, pdf, label=f'Composite Likelihood', alpha=1, linewidth=1)
        ax.set_xlabel('X')
        ax.set_ylabel('P(X)')
        ax.legend()
        if show:
            plt.show()
        return f, ax

    @checkfit
    def get_components(self) -> List[Dict[str, Union[bool, np.ndarray, float, int]]]:
        """
            get list of component parameters
        :return:
        """
        component_list = []
        if self.zero_inflated:
            component_list.append({'weight': self.weights[0], 'Zero Component': True})
        for i in range(int(self.zero_inflated), self.n_components):
            component_list.append({'weight': self.weights[i]} | {key: self.params[key][i-int(self.zero_inflated)] for key in self.params})
        return component_list


def initialize_weights_uniform(n_components):
    return np.ones(n_components) / n_components




import numpy as np
import scipy.stats
from typing import Tuple, Union
from ..base import MixtureModel
from ...utils.static import Params, eps


class ZeroInflatedPoissonMixtureModel(MixtureModel):
    _param_signature_ = {'mu': {'type': float, 'max': np.inf, 'min': 0},}
    _support = {'min': 0-eps, 'max': np.inf, 'dtype': int}
    pdf_thresholder = 10e-8

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = 'ZeroInflatedPoissonMixtureModel'
        self.zero_inflated = True
        self.skip_deterministic = True
        self.min_components = 2
        self.n_variables = 1
        self.discrete = True

    def likelihood(self, weights: np.ndarray, data: np.ndarray,
                   log=False, **params) -> Union[np.ndarray, Union[float, np.float64]]:

        likelihood_poisson = scipy.stats.poisson.pmf(data[:, 1:], **params)
        likelihood_zero = (data[:, 0] == 0).astype(np.float64)[:, np.newaxis]
        likelihood_weighted = np.concatenate((likelihood_zero, likelihood_poisson), axis=1) * weights

        if log:
            return self.log_likelihood(likelihood_weighted)
        else:
            return likelihood_weighted

    def _fit_deterministic(self, X):
        return self._fit_(X, self.min_components)

    def _fit_(self, X: np.ndarray, n_components: int) -> Tuple[Params, np.ndarray, float]:
        pi = (X == 0).mean()
        weights = np.hstack((pi, (1 - pi) * np.ones(n_components - 1) / (n_components - 1)))
        unique, counts = np.unique(X[X > 0], return_counts=True)
        try:
            rate = np.random.choice(unique, n_components - 1, replace=False,
                                    p=counts / counts.sum()).astype(np.float32)
        except ValueError:  # not enough variety
            rate = X[X != 0].mean() * (np.random.random(n_components - 1) / (n_components - 1) + .5)

        log_likelihoods = [-np.inf]
        X_rep = np.repeat(X[..., np.newaxis], n_components, axis=1)
        for iteration in range(self.max_iterations):
            # Expectation
            likelihood_weighted = self.likelihood(weights=weights, data=X_rep, log=False, mu=rate)
            log_likelihood = self.log_likelihood(likelihood_weighted)
            log_likelihoods.append(log_likelihood)
            # if abs(log_likelihoods[-1] - log_likelihoods[-2]) < 1e-2:
            #     break
            if self.check_convergence(log_likelihoods):
                break
            r = (eps + likelihood_weighted) / (likelihood_weighted + self.eps).sum(axis=1)[..., np.newaxis]

            # maximization
            rate = (r[:, 1:] * X_rep[:, 1:]).sum(axis=0) / (r[:, 1:] + self.eps).sum(axis=0)
            weights = r.sum(axis=0) / r.sum()

        return {'mu': rate}, weights, log_likelihoods[-1]

    def _generate_component(self, component, n):
        return scipy.stats.poisson.rvs(mu=self.params['mu'][component-1], size=n)

    def _cdf_separate(self, X):
        zero_component = (X[:, 0] <= 0).astype(float)[..., np.newaxis]
        non_zero = scipy.stats.poisson.cdf(X + self.eps, **self.params)
        return np.concatenate((zero_component, non_zero), axis=-1)

    def _sf_separate(self, X):
        zero_component = (X[:, 0] >= 0).astype(float)[..., np.newaxis]
        non_zero = scipy.stats.poisson.sf(X - self.eps, **self.params)
        return np.concatenate((zero_component, non_zero), axis=-1)

